import asyncio
from datetime import datetime, timedelta
from types import TracebackType
from typing import Union, Optional, Type, Dict, Any

import aiohttp

from .regions import APIRegion


class API:
    def __init__(self, client_id: str, client_secret: str, client_region: Union[APIRegion, str], *,
                 max_parallel_requests: Optional[int] = None, max_request_retries: Optional[int] = None,
                 request_retry_delay: Optional[int] = None, request_debugging: Optional[bool] = None):
        """A class with methods for interacting with Battle.net's various APIs

        :param client_id: Battle.net Project Client ID - Generated at https://develop.battle.net/access/clients
        :type client_id: str
        :param client_secret: Battle.net Project Client Secret - Generated at https://develop.battle.net/access/clients
        :type client_secret: str
        :param client_region: The Battle.net/WoW API region which we submit requests to, defaults to APIRegion.US
        :type client_region: str, APIRegion, optional
        :param max_parallel_requests: The maximum number of parallel aiohttp requests (Default: 50)
        :type max_parallel_requests: int, optional
        :param max_request_retries: The maximum number of aiohttp request retries (min 1) (Default: 3)
        :type max_request_retries: int, optional
        :param request_retry_delay: The delay between aiohttp request retries (Seconds - min 0)(Default: 1)
        :type request_retry_delay: int, optional
        :param request_debugging: Whether aiohttp request exceptions are raised or return None (Default: False)
        :type request_debugging: bool, optional
        """

        self.__client_id: str = client_id
        self.__client_secret: str = client_secret
        self.__client_region: APIRegion = self.set_region(client_region)
        self.__client_locale: str = self.__client_region.value['supported_locales'][0]

        self.__access_tokens: Dict[str, Dict[str, Any]] = {}

        self.__semaphore: asyncio.Semaphore = asyncio.Semaphore(
            max_parallel_requests if max_parallel_requests is not None else 50
        )

        self.__max_request_retries: int = max_request_retries if (
            max_request_retries is not None and max_request_retries > 1
        ) else 3

        self.__request_retry_delay: int = request_retry_delay if (
            request_retry_delay is not None and request_retry_delay > 0
        ) else 1

        self.__request_debugging: bool = request_debugging if (request_debugging is not None) else True

        self.__session: Optional[aiohttp.ClientSession] = None

        self.__is_context_manager: bool = False

    def __enter__(self) -> None:
        raise TypeError("Use 'async with' instead")

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        pass

    async def __aenter__(self) -> 'API':
        self.__session = aiohttp.ClientSession()
        self.__is_context_manager = True
        return self

    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> None:
        if self.__session is not None and self.__session.closed is False:
            await self.__session.close()

    def get_region(self) -> str:
        """Returns the current region being used for API requests

        :return: The current region being used for API requests
        :rtype: str
        """
        return self.__client_region.name

    def get_locale(self) -> str:
        """Returns the current locale being used for API requests

        :return: The current locale being used for API requests
        :rtype: str
        """
        return self.__client_locale

    def set_region(self, region: Union[APIRegion, str]) -> APIRegion:
        """Sets the region to be used for API requests

        :param region: The desired region to be used for API requests
        :type region: str, APIRegion
        :raises InvalidRegionException: Raised when the provided region isn't supported/found
        :return: The API object
        :rtype: API
        """
        if region in list(APIRegion) and isinstance(region, APIRegion):
            self.__client_region = region
            self.__client_locale = region.value['supported_locales'][0]
            return region
        elif isinstance(region, str):
            for i in list(APIRegion):
                if i.name.lower() == region.lower():
                    self.__client_region = i
                    self.__client_locale = i.value['supported_locales'][0]
                    return i
            raise InvalidRegionException('Invalid API Region {}, supported regions are {}'.format(
                region, list(APIRegion.__members__.keys())))
        else:
            raise InvalidRegionException('Invalid API Region {}, supported regions are {}'.format(
                region, list(APIRegion.__members__.keys())))

    def set_locale(self, locale: str) -> str:
        """Sets the locale we'll use for API requests

        :param locale: The desired locale to be used for API requests.
        :type locale: str
        :raises InvalidLocaleException: Raised the input doesn't match any of the current region's supported locales
        :return: The current locale being used for API requests
        :rtype: str
        """
        if locale and locale in self.__client_region.value['supported_locales']:
            self.__client_locale = locale
        else:
            raise InvalidLocaleException('Invalid Regional Locale {}, supported locales for {} are {}'.format(
                locale, self.__client_region.name, self.__client_region.value['supported_locales']))

        return self.__client_locale

    def get_hostname(self) -> str:
        """Returns the current region's hostname for Game API requests

        :return: The current region's hostname for Game API requests
        :rtype: str
        """
        return self.__client_region.value['game_api_hostname']

    def get_oauth_hostname(self) -> str:
        """Returns the current region's hostname for OAuth API requests

        :return: The current region's hostname for OAuth API requests
        :rtype: str
        """
        return self.__client_region.value['oauth_api_hostname']

    async def get_access_token(self) -> str:
        """Returns and / or generates an API access token using the provided credentials in the class constructor

        :raises RequestException: Raised when we encounter an issue when making an aiohttp request.
        :return: A Battle.net OAuth Access Token
        :rtype: str
        """

        if self.__client_region.name in self.__access_tokens and \
                self.__access_tokens[self.__client_region.name]['Expires'] > datetime.now():
            return self.__access_tokens[self.__client_region.name]['Token']
        else:
            endpoint = f"/oauth/token"

            hostname = self.get_oauth_hostname()

            params = {'grant_type': 'client_credentials'}

            data = await self.get_resource(hostname, endpoint, params,
                                           auth=aiohttp.BasicAuth(self.__client_id, self.__client_secret),
                                           method="POST")

            if data is None:
                raise AccessTokenException(
                    'Failed to retrieve an access token, verify your credentials & internet connectivity.')

            expires = datetime.now() + timedelta(seconds=data['expires_in'] - 60)

            self.__access_tokens[self.__client_region.name] = {'Token': data['access_token'], 'Expires': expires}

            return self.__access_tokens[self.__client_region.name]['Token']

    @staticmethod
    async def multi_request(requests: list) -> Optional[Union[tuple, list]]:
        """Make several API requests in parallel

        :param requests: A list of API request coroutines
        :type requests: list
        :return: The API responses as a list of dicts
        :rtype: list
        """
        data = await asyncio.gather(*requests)
        if data:
            return data
        else:
            return None

    async def get_resource(self,
                           hostname: str, api_endpoint: str,
                           params: Optional[dict] = None,
                           auth: Optional[aiohttp.BasicAuth] = None,
                           method: Optional[str] = None,
                           ) -> Optional[dict]:
        """Make an API request and return the response as a JSON dictionary

        :param hostname: The hostname to make the request to
        :type hostname: str
        :param api_endpoint: The API endpoint following the regional hostname we wish to send a request to.
        :type api_endpoint: str
        :param params: The additional arguments/parameters we need to send with the request, defaults to None
        :type params: dict, optional
        :param auth: The aiohttp BasicAuth object to use for authentication, defaults to None
        :type auth: aiohttp.BasicAuth, optional
        :param method: The HTTP method to use for the request, defaults to "GET"
        :type method: str, optional
        :raises RequestMethodException: Raised when an invalid HTTP request method is selected.
        :raises RequestException: Raised when we encounter an issue when making an aiohttp request.
        :return: The response from the API as a JSON dictionary
        :rtype: dict, none
        """

        if method is None:
            method = "GET"

        async with self.__semaphore:
            current_attempt: int = 1
            local_session: Optional[aiohttp.ClientSession] = aiohttp.ClientSession() if \
                (self.__is_context_manager) is False else None

            if self.__is_context_manager is True and (self.__session is None or self.__session.closed):
                self.__session = aiohttp.ClientSession()
            result: Optional[Dict] = None
            while current_attempt <= self.__max_request_retries and result is None:
                try:

                    supported_methods = {
                        "GET": local_session.get if (
                            self.__is_context_manager is False and local_session is not None
                        ) else self.__session.get,
                        "POST": local_session.post if (
                            self.__is_context_manager is False and local_session is not None
                        ) else self.__session.post,
                    }

                    if method.upper() not in supported_methods:
                        raise RequestMethodException('Invalid HTTP request method {}, supported methods are {}'.format(
                            method, list(supported_methods.keys())))

                    async with supported_methods[method](
                            hostname.format(api_endpoint=api_endpoint),
                            params=params,
                            auth=auth
                    ) as response:

                        if response.status == 200:
                            result = await response.json()

                        response.raise_for_status()

                    if local_session is not None and local_session.closed is False:
                        await local_session.close()

                except aiohttp.ClientError:
                    if current_attempt == self.__max_request_retries:
                        if self.__request_debugging:
                            raise
                        return result
                    current_attempt += 1
                    await asyncio.sleep(self.__request_retry_delay)
            return result


class ApiException(Exception):
    """Generic exception type for our API

    :param message: Description of the occurring error
    :type message: str, optional
    """

    def __init__(self, message: str = "An error was encountered with the API class."):
        super().__init__(message)


class AccessTokenException(ApiException):
    """Exception thrown when a null access token is provided

    :param message: Description of the occurring error
    :type message: str, optional
    """

    def __init__(self, message: str = "An error was encountered retrieving an API access token"):
        super().__init__(message)


class RequestException(ApiException):
    """Exception thrown when something goes wrong with an HTTP API request

    :param message: Description of the occurring error
    :type message: str, optional
    """

    def __init__(self, message: str = "Error making API request"):
        super().__init__(message)


class RequestMethodException(ApiException):
    """Exception thrown when an invalid request method is given

    :param message: Description of the occurring error
    :type message: str, optional
    """

    def __init__(self, message: str = "Invalid request method"):
        super().__init__(message)


class InvalidRegionException(ApiException):
    """Exception thrown when an invalid/unsupported API Region is provided

    :param message: Description of the occurring error
    :type message: str, optional
    """

    def __init__(self, message: str = "Invalid API Region"):
        super().__init__(message)


class InvalidLocaleException(ApiException):
    """Exception thrown when an invalid/unsupported API Locale is provided

    :param message: Description of the occurring error
    :type message: str, optional
    """

    def __init__(self, message: str = "Invalid API Locale"):
        super().__init__(message)



from .game_data_classic import GameData


class ClassicApi:
    def __init__(self, api):
        """This class adds structure to our wrapper, creating accessible API endpoint methods for classic World of Warcraft
        :param api: An instance of our generic API object
        :type api: API
        """
        self.GameData = GameData(api)



from enum import Enum


class APIRegion(Enum):
    US: dict = {
        'oauth_api_hostname': 'https://us.battle.net{api_endpoint}',
        'game_api_hostname': 'https://us.api.blizzard.com{api_endpoint}',
        'supported_locales': ['en_US', 'es_MX', 'pt_BR']
    }
    EU: dict = {
        'oauth_api_hostname': 'https://eu.battle.net{api_endpoint}',
        'game_api_hostname': 'https://eu.api.blizzard.com{api_endpoint}',
        'supported_locales': ['en_GB', 'es_ES', 'fr_FR', 'ru_RU', 'de_DE', 'pt_PT', 'it_IT']
    }
    KR: dict = {
        'oauth_api_hostname': 'https://kr.battle.net{api_endpoint}',
        'game_api_hostname': 'https://kr.api.blizzard.com{api_endpoint}',
        'supported_locales': ['ko_KR']
    }
    TW: dict = {
        'oauth_api_hostname': 'https://tw.battle.net{api_endpoint}',
        'game_api_hostname': 'https://tw.api.blizzard.com{api_endpoint}',
        'supported_locales': ['zh_TW']
    }
    CN: dict = {
        'oauth_api_hostname': 'https://www.battlenet.com.cn{api_endpoint}',
        'game_api_hostname': 'https://gateway.battlenet.com.cn{api_endpoint}',
        'supported_locales': ['zh_CN']
    }



"""
World of Warcraft API Wrapper
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

An asynchronous wrapper for the World of Warcraft API

:copyright: (c) 2021-Present Adalyia
:license: MIT, see LICENSE for more details.

"""

from .api import *
from .regions import *
from .wowapi import *



from typing import Optional


class GameData:
    def __init__(self, api):
        """This class contains all API endpoints for the GameData category of the Classic World of Warcraft API

        :param api: An instance of our generic API object
        :type api: API
        """
        self.api = api

    async def get_game_api_resource(self, namespace: str, endpoint: str, params: dict = None) -> Optional[dict]:
        """Generic method for retrieving data from a Classic Game Data API endpoint

        :param namespace: The namespace of the resource we're trying to access
        :type namespace: str
        :param endpoint: The API endpoint of the resource we're trying to access
        :type endpoint: str
        :param params: Parameters to send with the request, defaults to None
        :type params: dict, optional
        :return: The result of the API request (Warning: Can be None/Null)
        :rtype: dict
        """
        region = self.api.get_region()
        locale = self.api.get_locale()
        hostname = self.api.get_hostname()
        token = await self.api.get_access_token()

        if params is None:
            params = {}

        params["namespace"] = namespace.format(region=region),
        params["locale"] = locale,
        params["access_token"] = token

        return await self.api.get_resource(hostname, endpoint, params)

    # region Auction House API

    async def get_auction_house_index(self, connected_realm_id: int):
        """Returns an index of auction houses for a connected realm.

See the Connected Realm API for information about retrieving a list of connected realm IDs.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :return: Returns an index of auction houses for a connected realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}/auctions/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_auctions(self, connected_realm_id: int, auction_house_id: int):
        """Returns all active auctions for a specific auction house on a connected realm.

See the Connected Realm API for information about retrieving a list of connected realm IDs.

Auction house data updates at a set interval. The value was initially set at 1 hour; however, it might change over time without notice.

Depending on the number of active auctions on the specified connected realm, the response from this endpoint may be rather large, sometimes exceeding 10 MB.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :param auction_house_id: The ID of the auction house.
        :type auction_house_id: int
        :return: Returns all active auctions for a specific auction house on a connected realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}/auctions/{auction_house_id}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Connected Realm API

    async def get_connected_realms_index(self):
        """Returns an index of connected realms.

        :return: Returns an index of connected realms.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_connected_realm(self, connected_realm_id: int):
        """Returns a connected realm by ID.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :return: Returns a connected realm by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_connected_realms_search(self, search_params: Optional[dict] = None):
        """Performs a search of connected realms. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of connected realms. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/connected-realm"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Creature API

    async def get_creature_families_index(self):
        """Returns an index of creature families.

        :return: Returns an index of creature families.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-family/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_family(self, creature_family_id: int):
        """Returns a creature family by ID.

        :param creature_family_id: The ID of the creature family.
        :type creature_family_id: int
        :return: Returns a creature family by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-family/{creature_family_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_types_index(self):
        """Returns an index of creature types.

        :return: Returns an index of creature types.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-type/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_type(self, creature_type_id: int):
        """Returns a creature type by ID.

        :param creature_type_id: The ID of the creature type.
        :type creature_type_id: int
        :return: Returns a creature type by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-type/{creature_type_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature(self, creature_id: int):
        """Returns a creature by ID.

        :param creature_id: The ID of the creature.
        :type creature_id: int
        :return: Returns a creature by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature/{creature_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_search(self, search_params: Optional[dict] = None):
        """Performs a search of creatures. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of creatures. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/creature"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    async def get_creature_display_media(self, creature_display_id: int):
        """Returns media for a creature display by ID.

        :param creature_display_id: The ID of the creature display.
        :type creature_display_id: int
        :return: Returns media for a creature display by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/creature-display/{creature_display_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_family_media(self, creature_family_id: int):
        """Returns media for a creature family by ID.

        :param creature_family_id: The ID of the creature family.
        :type creature_family_id: int
        :return: Returns media for a creature family by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/creature-family/{creature_family_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Guild Crest API

    async def get_guild_crest_components_index(self):
        """Returns an index of guild crest media.

        :return: Returns an index of guild crest media.
        :rtype: dict
        """
        endpoint = f"/data/wow/guild-crest/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_guild_crest_border_media(self, border_id: int):
        """Returns media for a guild crest border by ID.

        :param border_id: The ID of the guild crest border.
        :type border_id: int
        :return: Returns media for a guild crest border by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/guild-crest/border/{border_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_guild_crest_emblem_media(self, emblem_id: int):
        """Returns media for a guild crest emblem by ID.

        :param emblem_id: The ID of the guild crest emblem.
        :type emblem_id: int
        :return: Returns media for a guild crest emblem by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/guild-crest/emblem/{emblem_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Item API

    async def get_item_classes_index(self):
        """Returns an index of item classes.

        :return: Returns an index of item classes.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-class/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_class(self, item_class_id: str):
        """Returns an item class by ID.

        :param item_class_id: The ID of the item class.
        :type item_class_id: str
        :return: Returns an item class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-class/{item_class_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_subclass(self, item_class_id: str, item_subclass_id: str):
        """Returns an item subclass by ID.

        :param item_class_id: The ID of the item class.
        :type item_class_id: str
        :param item_subclass_id: The ID of the item subclass.
        :type item_subclass_id: str
        :return: Returns an item subclass by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-class/{item_class_id}/item-subclass/{item_subclass_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item(self, item_id: str):
        """Returns an item by ID.

        :param item_id: The ID of the item.
        :type item_id: str
        :return: Returns an item by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item/{item_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_media(self, item_id: int):
        """Returns media for an item by ID.

        :param item_id: The ID of the item.
        :type item_id: int
        :return: Returns media for an item by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/item/{item_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_search(self, search_params: Optional[dict] = None):
        """Performs a search of items. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of items. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/item"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Media Search API

    async def get_media_search(self, search_params: Optional[dict] = None):
        """Performs a search of all types of media documents. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of all types of media documents. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/media"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Playable Class API

    async def get_playable_classes_index(self):
        """Returns an index of playable classes.

        :return: Returns an index of playable classes.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-class/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_class(self, class_id: int):
        """Returns a playable class by ID.

        :param class_id: The ID of the playable class.
        :type class_id: int
        :return: Returns a playable class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-class/{class_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_class_media(self, playable_class_id: int):
        """Returns media for a playable class by ID.

        :param playable_class_id: The ID of the playable class.
        :type playable_class_id: int
        :return: Returns media for a playable class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/playable-class/{playable_class_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Playable Race API

    async def get_playable_races_index(self):
        """Returns an index of playable races.

        :return: Returns an index of playable races.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-race/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_race(self, playable_race_id: int):
        """Returns a playable race by ID.

        :param playable_race_id: The ID of the playable race.
        :type playable_race_id: int
        :return: Returns a playable race by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-race/{playable_race_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Power Type API

    async def get_power_types_index(self):
        """Returns an index of power types.

        :return: Returns an index of power types.
        :rtype: dict
        """
        endpoint = f"/data/wow/power-type/index"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_power_type(self, power_type_id: int):
        """Returns a power type by ID.

        :param power_type_id: The ID of the power type.
        :type power_type_id: int
        :return: Returns a power type by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/power-type/{power_type_id}"
        namespace = "static-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region PvP Season API

    async def get_pvp_seasons_index(self):
        """Returns an index of PvP seasons.

        :return: Returns an index of PvP seasons.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_season(self, pvp_season_id: int):
        """Returns a PvP season by ID.

        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns a PvP season by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/{pvp_season_id}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_region_index(self):
        """Returns an index of PvP Regions.

        :return: Returns an index of PvP Regions.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-region/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_regional_season_index(self, pvp_region_id: int):
        """Returns an index of PvP Seasons in a PvP region.

        :param pvp_region_id: The ID of the PvP region.
        :type pvp_region_id: int
        :return: Returns an index of PvP Seasons in a PvP region.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-region/{pvp_region_id}/pvp-season/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_regional_season(self, pvp_region_id: int, pvp_season_id: int):
        """Returns a PvP season by region ID and season ID.

        :param pvp_region_id: The ID of the PvP region.
        :type pvp_region_id: int
        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns a PvP season by region ID and season ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-region/{pvp_region_id}/pvp-season/{pvp_season_id}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_leaderboards_index(self, pvp_region_id: int, pvp_season_id: int):
        """Returns an index of PvP leaderboards for a PvP season in a given PvP region.

        :param pvp_region_id: The ID of the PvP region.
        :type pvp_region_id: int
        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns an index of PvP leaderboards for a PvP season in a given PvP region.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-region/{pvp_region_id}/pvp-season/{pvp_season_id}/pvp-leaderboard/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_leaderboard(self, pvp_region_id: int, pvp_season_id: int, pvp_bracket: str):
        """Returns the PvP leaderboard of a specific PvP bracket for a PvP season in a given PvP region.

        :param pvp_region_id: The ID of the PvP region.
        :type pvp_region_id: int
        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :param pvp_bracket: The PvP bracket type.
        :type pvp_bracket: str
        :return: Returns the PvP leaderboard of a specific PvP bracket for a PvP season in a given PvP region.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-region/{pvp_region_id}/pvp-season/{pvp_season_id}/pvp-leaderboard/{pvp_bracket}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_rewards_index(self, pvp_region_id: int, pvp_season_id: int):
        """Returns an index of PvP rewards for a PvP season in a given PvP region.

        :param pvp_region_id: The ID of the PvP region.
        :type pvp_region_id: int
        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns an index of PvP rewards for a PvP season in a given PvP region.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-region/{pvp_region_id}/pvp-season/{pvp_season_id}/pvp-reward/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Realm API

    async def get_realms_index(self):
        """Returns an index of realms.

        :return: Returns an index of realms.
        :rtype: dict
        """
        endpoint = f"/data/wow/realm/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_realm(self, realm_slug: str):
        """Returns a single realm by slug or ID.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :return: Returns a single realm by slug or ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/realm/{realm_slug}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_realm_search(self, search_params: Optional[dict] = None):
        """Performs a search of realms. The fields below are examples only. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of realms. The fields below are examples only. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/realm"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Region API

    async def get_regions_index(self):
        """Returns an index of regions.

        :return: Returns an index of regions.
        :rtype: dict
        """
        endpoint = f"/data/wow/region/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_region(self, region_id: int):
        """Returns a region by ID.

        :param region_id: The ID of the region.
        :type region_id: int
        :return: Returns a region by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/region/{region_id}"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region WoW Token API

    async def get_wow_token_index_cn(self):
        """Returns the WoW Token index.

        :return: Returns the WoW Token index.
        :rtype: dict
        """
        endpoint = f"/data/wow/token/index"
        namespace = "dynamic-classic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion



from typing import Optional


class Profile:
    def __init__(self, api):
        """This class contains all API endpoints for the Profile category of the Retail World of Warcraft API

        :param api: An instance of our generic API object
        :type api: API
        """
        self.api = api

    async def get_profile_api_resource(self, namespace: str, endpoint: str, params: dict = None) -> Optional[dict]:
        """Generic method for retrieving data from a Profile API endpoint

        :param namespace: The namespace of the resource we're trying to access
        :type namespace: str
        :param endpoint: The API endpoint of the resource we're trying to access
        :type endpoint: str
        :param params: Parameters to send with the request, defaults to None
        :type params: dict, optional
        :return: The result of the API request (Warning: Can be None/Null)
        :rtype: dict
        """
        region = self.api.get_region()
        locale = self.api.get_locale()
        hostname = self.api.get_hostname()
        token = await self.api.get_access_token()

        if params is None:
            params = {}

        params["namespace"] = namespace.format(region=region),
        params["locale"] = locale,
        params["access_token"] = token

        return await self.api.get_resource(hostname, endpoint, params)

    # region Character Achievements API

    async def get_character_achievements_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of the achievements a character has completed.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of the achievements a character has completed.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/achievements"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_achievement_statistics(self, realm_slug: str, character_name: str):
        """Returns a character's statistics as they pertain to achievements.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a character's statistics as they pertain to achievements.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/achievements/statistics"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Appearance API

    async def get_character_appearance_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of a character's appearance settings.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of a character's appearance settings.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/appearance"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Collections API

    async def get_character_collections_index(self, realm_slug: str, character_name: str):
        """Returns an index of collection types for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns an index of collection types for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/collections"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_mounts_collection_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of the mounts a character has obtained.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of the mounts a character has obtained.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/collections/mounts"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_pets_collection_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of the battle pets a character has obtained.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of the battle pets a character has obtained.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/collections/pets"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Encounters API

    async def get_character_encounters_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of a character's encounters.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of a character's encounters.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/encounters"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_dungeons(self, realm_slug: str, character_name: str):
        """Returns a summary of a character's completed dungeons.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of a character's completed dungeons.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/encounters/dungeons"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_raids(self, realm_slug: str, character_name: str):
        """Returns a summary of a character's completed raids.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of a character's completed raids.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/encounters/raids"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Equipment API

    async def get_character_equipment_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of the items equipped by a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of the items equipped by a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/equipment"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Hunter Pets API

    async def get_character_hunter_pets_summary(self, realm_slug: str, character_name: str):
        """If the character is a hunter, returns a summary of the character's hunter pets. Otherwise, returns an HTTP 404 Not Found error.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: If the character is a hunter, returns a summary of the character's hunter pets. Otherwise, returns an HTTP 404 Not Found error.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/hunter-pets"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Media API

    async def get_character_media_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of the media assets available for a character (such as an avatar render).

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of the media assets available for a character (such as an avatar render).
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/character-media"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Mythic Keystone Profile API

    async def get_character_mythic_keystone_profile_index(self, realm_slug: str, character_name: str):
        """Returns the Mythic Keystone profile index for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns the Mythic Keystone profile index for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/mythic-keystone-profile"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_mythic_keystone_season_details(self, realm_slug: str, character_name: str, season_id: str):
        """Returns the Mythic Keystone season details for a character.

Returns a 404 Not Found for characters that have not yet completed a Mythic Keystone dungeon for the specified season.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :param season_id: The ID of the Mythic Keystone season.
        :type season_id: str
        :return: Returns the Mythic Keystone season details for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/mythic-keystone-profile/season/{season_id}"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Professions API

    async def get_character_professions_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of professions for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of professions for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/professions"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Profile API

    async def get_character_profile_summary(self, realm_slug: str, character_name: str):
        """Returns a profile summary for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a profile summary for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_profile_status(self, realm_slug: str, character_name: str):
        """Returns the status and a unique ID for a character. A client should delete information about a character from their application if any of the following conditions occur:
an HTTP 404 Not Found error is returned
the is_valid value is false
the returned character ID doesn't match the previously recorded value for the character
The following example illustrates how to use this endpoint:
A client requests and stores information about a character, including its unique character ID and the timestamp of the request.
After 30 days, the client makes a request to the status endpoint to verify if the character information is still valid.
If character cannot be found, is not valid, or the characters IDs do not match, the client removes the information from their application.
If the character is valid and the character IDs match, the client retains the data for another 30 days.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns the status and a unique ID for a character. A client should delete information about a character from their application if any of the following conditions occur:
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/status"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character PvP API

    async def get_character_pvp_bracket_statistics(self, realm_slug: str, character_name: str, pvp_bracket: str):
        """Returns the PvP bracket statistics for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :param pvp_bracket: The PvP bracket type.
        :type pvp_bracket: str
        :return: Returns the PvP bracket statistics for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/pvp-bracket/{pvp_bracket}"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_pvp_summary(self, realm_slug: str, character_name: str):
        """Returns a PvP summary for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a PvP summary for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/pvp-summary"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Quests API

    async def get_character_quests(self, realm_slug: str, character_name: str):
        """Returns a character's active quests as well as a link to the character's completed quests.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a character's active quests as well as a link to the character's completed quests.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/quests"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_character_completed_quests(self, realm_slug: str, character_name: str):
        """Returns a list of quests that a character has completed.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a list of quests that a character has completed.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/quests/completed"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Reputations API

    async def get_character_reputations_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of a character's reputations.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of a character's reputations.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/reputations"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Soulbinds API

    async def get_character_soulbinds(self, realm_slug: str, character_name: str):
        """Returns a character's soulbinds.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a character's soulbinds.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/soulbinds"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Specializations API

    async def get_character_specializations_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of a character's specializations.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of a character's specializations.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/specializations"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Statistics API

    async def get_character_statistics_summary(self, realm_slug: str, character_name: str):
        """Returns a statistics summary for a character.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a statistics summary for a character.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/statistics"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Character Titles API

    async def get_character_titles_summary(self, realm_slug: str, character_name: str):
        """Returns a summary of titles a character has obtained.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param character_name: The lowercase name of the character.
        :type character_name: str
        :return: Returns a summary of titles a character has obtained.
        :rtype: dict
        """
        endpoint = f"/profile/wow/character/{realm_slug}/{character_name}/titles"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion
    # region Guild API

    async def get_guild(self, realm_slug: str, name_slug: str):
        """Returns a single guild by its name and realm.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param name_slug: The slug of the guild.
        :type name_slug: str
        :return: Returns a single guild by its name and realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/guild/{realm_slug}/{name_slug}"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_guild_activity(self, realm_slug: str, name_slug: str):
        """Returns a single guild's activity by name and realm.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param name_slug: The slug of the guild.
        :type name_slug: str
        :return: Returns a single guild's activity by name and realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/guild/{realm_slug}/{name_slug}/activity"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_guild_achievements(self, realm_slug: str, name_slug: str):
        """Returns a single guild's achievements by name and realm.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param name_slug: The slug of the guild.
        :type name_slug: str
        :return: Returns a single guild's achievements by name and realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/guild/{realm_slug}/{name_slug}/achievements"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    async def get_guild_roster(self, realm_slug: str, name_slug: str):
        """Returns a single guild's roster by its name and realm.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :param name_slug: The slug of the guild.
        :type name_slug: str
        :return: Returns a single guild's roster by its name and realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/guild/{realm_slug}/{name_slug}/roster"
        namespace = "profile-{region}"

        return await self.get_profile_api_resource(namespace, endpoint)

    # endregion



"""
Classic World of Warcraft API Endpoints
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Module containing endpoint access methods for the Classic World of Warcraft API.

:copyright: (c) 2021-Present Adalyia
:license: MIT, see LICENSE for more details.

"""

from .game_data_classic import *



import re
from typing import Dict, Optional
from urllib.parse import unquote

from . import API
from .classic.classic import ClassicApi
from .retail.retail import RetailApi


class WowApi(API):
    def __init__(self, *args, **kwargs):
        """This class contains some useful functions/QoL features for working with the World of Warcraft API.
        For additional arguments see the API class documentation.
        """
        super().__init__(*args, **kwargs)

        self.Retail = RetailApi(super())
        self.Classic = ClassicApi(super())

        self.__realms = None

    @staticmethod
    async def parse_armory_link(url: str) -> Optional[Dict[str, str]]:
        """Parses a World of Warcraft Armoury link and returns the character's name, realm, and region

        :param url: A WoW armoury link (ex: https://worldofwarcraft.com/en-us/character/us/{slug}/{character})
        :type url: str
        :return: {'name': name, 'realm': realm_slug, 'region': region}
        :rtype: dict
        """
        string = unquote(str(url))

        found = re.search(
            r"/(us|eu|kr|tw|cn)/([0-9\-\w]*)/([a-zA-Z\w]*)", string, re.U
        )

        if found and found.group() and len(found.groups()) == 3:
            name = found.group(3).lower()
            realm_slug = found.group(2).lower()
            region = found.group(1).lower()

            return {'name': name, 'realm': realm_slug, 'region': region}

        return None

    async def get_realm_slug(self, realm_name: str) -> Optional[str]:
        """Attempts to match user input with a WoW realm and return its slug

        :param realm_name: A string to query the realms index with (full name, id, short name, etc)
        :type realm_name: str
        :return: A matching realm's slug
        :rtype: str
        """
        if self.__realms:
            for realm in self.__realms:
                realm_name = str(realm_name).lower()
                if realm_name in str(realm['name']).lower() or realm_name in str(
                        realm['fixed']).lower() or realm_name in str(
                        realm['slug']).lower() or realm_name == str(realm['id']).lower():
                    return realm['slug']
        else:
            data = await self.Retail.GameData.get_realms_index()

            if data:
                self.__realms = []
                for realm in data['realms']:
                    self.__realms.append(
                        {'name': realm['name'], 'fixed': str(realm['name']).replace(" ", ""), 'slug': realm['slug'],
                         'id': realm['id']})
                return await self.get_realm_slug(realm_name)

        return None

    @staticmethod
    async def format_wow_gold(money: int) -> str:
        """Converts a WoW money value to a formatted string of Gold, Silver, and Copper

        :param money: A WoW currency/money value in copper
        :type money: int
        :return: {}g {}s {}c
        :rtype: str
        """

        gold = int(money / 10000)
        silver = int((money % 10000) / 100)
        copper = int((money % 10000) % 100)

        return f'{gold:,}g {silver:,}s {copper:,}c'



from .game_data import GameData
from .profile import Profile


class RetailApi:
    def __init__(self, api):
        """This class just helps structure our wrapper, creates accessible API endpoint accessors for retail World of Warcraft

        :param api: An instance of our generic API object
        :type api: API
        """
        self.GameData = GameData(api)
        self.Profile = Profile(api)



from typing import Optional


class GameData:

    def __init__(self, api):
        """This class contains all API endpoints for the GameData category of the Retail World of Warcraft API

        :param api: An instance of our generic API object
        :type api: API
        """
        self.api = api

    async def get_game_api_resource(self, namespace: str, endpoint: str, params: dict = None) -> Optional[dict]:
        """Generic method for retrieving data from a Game Data API endpoint

        :param namespace: The namespace of the resource we're trying to access
        :type namespace: str
        :param endpoint: The API endpoint of the resource we're trying to access
        :type endpoint: str
        :param params: Parameters to send with the request, defaults to None
        :type params: dict, optional
        :return: The result of the API request (Warning: Can be None/Null)
        :rtype: dict
        """
        region = self.api.get_region()
        locale = self.api.get_locale()
        hostname = self.api.get_hostname()
        token = await self.api.get_access_token()

        if params is None:
            params = {}

        params["namespace"] = namespace.format(region=region),
        params["locale"] = locale,
        params["access_token"] = token

        return await self.api.get_resource(hostname, endpoint, params)

    # For some reason this API endpoint isn't documented on the Blizzard API website, so for now it's added manually
    async def get_commodity_auctions(self):
        """Returns all active commodity auctions for a region.

Auction house data updates at a set interval. The value was initially set at 1 hour; however, it might change over time without notice.

Depending on the number of active auctions on the region, the response from this endpoint may be rather large, sometimes exceeding 10 MB.

        :return: Returns all active commodity auctions for a region.
        :rtype: dict
        """
        endpoint = f"/data/wow/auctions/commodities"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # region Achievement API

    async def get_achievement_categories_index(self):
        """Returns an index of achievement categories.

        :return: Returns an index of achievement categories.
        :rtype: dict
        """
        endpoint = f"/data/wow/achievement-category/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_achievement_category(self, achievement_category_id: int):
        """Returns an achievement category by ID.

        :param achievement_category_id: The ID of the achievement category.
        :type achievement_category_id: int
        :return: Returns an achievement category by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/achievement-category/{achievement_category_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_achievements_index(self):
        """Returns an index of achievements.

        :return: Returns an index of achievements.
        :rtype: dict
        """
        endpoint = f"/data/wow/achievement/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_achievement(self, achievement_id: int):
        """Returns an achievement by ID.

        :param achievement_id: The ID of the achievement.
        :type achievement_id: int
        :return: Returns an achievement by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/achievement/{achievement_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_achievement_media(self, achievement_id: int):
        """Returns media for an achievement by ID.

        :param achievement_id: The ID of the achievement.
        :type achievement_id: int
        :return: Returns media for an achievement by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/achievement/{achievement_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Auction House API

    async def get_auctions(self, connected_realm_id: int):
        """Returns all active auctions for a connected realm.

See the Connected Realm API for information about retrieving a list of connected realm IDs.

Auction house data updates at a set interval. The value was initially set at 1 hour; however, it might change over time without notice.

Depending on the number of active auctions on the specified connected realm, the response from this endpoint may be rather large, sometimes exceeding 10 MB.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :return: Returns all active auctions for a connected realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}/auctions"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Azerite Essence API

    async def get_azerite_essences_index(self):
        """Returns an index of azerite essences.

        :return: Returns an index of azerite essences.
        :rtype: dict
        """
        endpoint = f"/data/wow/azerite-essence/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_azerite_essence(self, azerite_essence_id: str):
        """Returns an azerite essence by ID.

        :param azerite_essence_id: The ID of the azerite essence.
        :type azerite_essence_id: str
        :return: Returns an azerite essence by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/azerite-essence/{azerite_essence_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_azerite_essence_search(self, search_params: Optional[dict] = None):
        """Performs a search of azerite essences. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of azerite essences. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/azerite-essence"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    async def get_azerite_essence_media(self, azerite_essence_id: int):
        """Returns media for an azerite essence by ID.

        :param azerite_essence_id: The ID of the azerite essence.
        :type azerite_essence_id: int
        :return: Returns media for an azerite essence by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/azerite-essence/{azerite_essence_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Connected Realm API

    async def get_connected_realms_index(self):
        """Returns an index of connected realms.

        :return: Returns an index of connected realms.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_connected_realm(self, connected_realm_id: int):
        """Returns a connected realm by ID.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :return: Returns a connected realm by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_connected_realms_search(self, search_params: Optional[dict] = None):
        """Performs a search of connected realms. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of connected realms. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/connected-realm"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Covenant API

    async def get_covenant_index(self):
        """Returns an index of covenants.

        :return: Returns an index of covenants.
        :rtype: dict
        """
        endpoint = f"/data/wow/covenant/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_covenant(self, covenant_id: int):
        """Returns a covenant by ID.

        :param covenant_id: The ID of the covenant.
        :type covenant_id: int
        :return: Returns a covenant by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/covenant/{covenant_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_covenant_media(self, covenant_id: int):
        """Returns media for a covenant by ID.

        :param covenant_id: The ID of the covenant.
        :type covenant_id: int
        :return: Returns media for a covenant by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/covenant/{covenant_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_soulbind_index(self):
        """Returns an index of soulbinds.

        :return: Returns an index of soulbinds.
        :rtype: dict
        """
        endpoint = f"/data/wow/covenant/soulbind/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_soulbind(self, soulbind_id: int):
        """Returns a soulbind by ID.

        :param soulbind_id: The ID of the soulbind.
        :type soulbind_id: int
        :return: Returns a soulbind by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/covenant/soulbind/{soulbind_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_conduit_index(self):
        """Returns an index of conduits.

        :return: Returns an index of conduits.
        :rtype: dict
        """
        endpoint = f"/data/wow/covenant/conduit/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_conduit(self, conduit_id: int):
        """Returns a conduit by ID.

        :param conduit_id: The ID of the conduit.
        :type conduit_id: int
        :return: Returns a conduit by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/covenant/conduit/{conduit_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Creature API

    async def get_creature_families_index(self):
        """Returns an index of creature families.

        :return: Returns an index of creature families.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-family/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_family(self, creature_family_id: int):
        """Returns a creature family by ID.

        :param creature_family_id: The ID of the creature family.
        :type creature_family_id: int
        :return: Returns a creature family by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-family/{creature_family_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_types_index(self):
        """Returns an index of creature types.

        :return: Returns an index of creature types.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-type/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_type(self, creature_type_id: int):
        """Returns a creature type by ID.

        :param creature_type_id: The ID of the creature type.
        :type creature_type_id: int
        :return: Returns a creature type by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature-type/{creature_type_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature(self, creature_id: int):
        """Returns a creature by ID.

        :param creature_id: The ID of the creature.
        :type creature_id: int
        :return: Returns a creature by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/creature/{creature_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_search(self, search_params: Optional[dict] = None):
        """Performs a search of creatures. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of creatures. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/creature"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    async def get_creature_display_media(self, creature_display_id: int):
        """Returns media for a creature display by ID.

        :param creature_display_id: The ID of the creature display.
        :type creature_display_id: int
        :return: Returns media for a creature display by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/creature-display/{creature_display_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_creature_family_media(self, creature_family_id: int):
        """Returns media for a creature family by ID.

        :param creature_family_id: The ID of the creature family.
        :type creature_family_id: int
        :return: Returns media for a creature family by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/creature-family/{creature_family_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Guild Crest API

    async def get_guild_crest_components_index(self):
        """Returns an index of guild crest media.

        :return: Returns an index of guild crest media.
        :rtype: dict
        """
        endpoint = f"/data/wow/guild-crest/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_guild_crest_border_media(self, border_id: int):
        """Returns media for a guild crest border by ID.

        :param border_id: The ID of the guild crest border.
        :type border_id: int
        :return: Returns media for a guild crest border by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/guild-crest/border/{border_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_guild_crest_emblem_media(self, emblem_id: int):
        """Returns media for a guild crest emblem by ID.

        :param emblem_id: The ID of the guild crest emblem.
        :type emblem_id: int
        :return: Returns media for a guild crest emblem by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/guild-crest/emblem/{emblem_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Item API

    async def get_item_classes_index(self):
        """Returns an index of item classes.

        :return: Returns an index of item classes.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-class/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_class(self, item_class_id: str):
        """Returns an item class by ID.

        :param item_class_id: The ID of the item class.
        :type item_class_id: str
        :return: Returns an item class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-class/{item_class_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_sets_index(self):
        """Returns an index of item sets.

        :return: Returns an index of item sets.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-set/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_set(self, item_set_id: int):
        """Returns an item set by ID.

        :param item_set_id: The ID of the item set.
        :type item_set_id: int
        :return: Returns an item set by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-set/{item_set_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_subclass(self, item_class_id: str, item_subclass_id: str):
        """Returns an item subclass by ID.

        :param item_class_id: The ID of the item class.
        :type item_class_id: str
        :param item_subclass_id: The ID of the item subclass.
        :type item_subclass_id: str
        :return: Returns an item subclass by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item-class/{item_class_id}/item-subclass/{item_subclass_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item(self, item_id: str):
        """Returns an item by ID.

        :param item_id: The ID of the item.
        :type item_id: str
        :return: Returns an item by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/item/{item_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_media(self, item_id: int):
        """Returns media for an item by ID.

        :param item_id: The ID of the item.
        :type item_id: int
        :return: Returns media for an item by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/item/{item_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_item_search(self, search_params: Optional[dict] = None):
        """Performs a search of items. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of items. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/item"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Journal API

    async def get_journal_expansions_index(self):
        """Returns an index of journal expansions.

        :return: Returns an index of journal expansions.
        :rtype: dict
        """
        endpoint = f"/data/wow/journal-expansion/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_journal_expansion(self, journal_expansion_id: int):
        """Returns a journal expansion by ID.

        :param journal_expansion_id: The ID of the journal expansion.
        :type journal_expansion_id: int
        :return: Returns a journal expansion by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/journal-expansion/{journal_expansion_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_journal_encounters_index(self):
        """Returns an index of journal encounters.

        :return: Returns an index of journal encounters.
        :rtype: dict
        """
        endpoint = f"/data/wow/journal-encounter/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_journal_encounter(self, journal_encounter_id: int):
        """Returns a journal encounter by ID.

        :param journal_encounter_id: The ID of the journal encounter.
        :type journal_encounter_id: int
        :return: Returns a journal encounter by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/journal-encounter/{journal_encounter_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_journal_encounter_search(self, search_params: Optional[dict] = None):
        """Performs a search of journal encounters. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of journal encounters. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/journal-encounter"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    async def get_journal_instances_index(self):
        """Returns an index of journal instances.

        :return: Returns an index of journal instances.
        :rtype: dict
        """
        endpoint = f"/data/wow/journal-instance/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_journal_instance(self, journal_instance_id: int):
        """Returns a journal instance.

        :param journal_instance_id: The ID of the journal instance.
        :type journal_instance_id: int
        :return: Returns a journal instance.
        :rtype: dict
        """
        endpoint = f"/data/wow/journal-instance/{journal_instance_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_journal_instance_media(self, journal_instance_id: int):
        """Returns media for a journal instance by ID.

        :param journal_instance_id: The ID of the journal instance.
        :type journal_instance_id: int
        :return: Returns media for a journal instance by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/journal-instance/{journal_instance_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Media Search API

    async def get_media_search(self, search_params: Optional[dict] = None):
        """Performs a search of all types of media documents. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of all types of media documents. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/media"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Modified Crafting API

    async def get_modified_crafting_index(self):
        """Returns the parent index for Modified Crafting.

        :return: Returns the parent index for Modified Crafting.
        :rtype: dict
        """
        endpoint = f"/data/wow/modified-crafting/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_modified_crafting_category_index(self):
        """Returns the index of Modified Crafting categories.

        :return: Returns the index of Modified Crafting categories.
        :rtype: dict
        """
        endpoint = f"/data/wow/modified-crafting/category/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_modified_crafting_category(self, category_id: int):
        """Returns a Modified Crafting category by ID.

        :param category_id: The ID of the Modified Crafting category.
        :type category_id: int
        :return: Returns a Modified Crafting category by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/modified-crafting/category/{category_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_modified_crafting_reagent_slot_type_index(self):
        """Returns the index of Modified Crafting reagent slot types.

        :return: Returns the index of Modified Crafting reagent slot types.
        :rtype: dict
        """
        endpoint = f"/data/wow/modified-crafting/reagent-slot-type/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_modified_crafting_reagent_slot_type(self, slot_type_id: int):
        """Returns a Modified Crafting reagent slot type by ID.

        :param slot_type_id: The ID of the Modified Crafting reagent slot type.
        :type slot_type_id: int
        :return: Returns a Modified Crafting reagent slot type by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/modified-crafting/reagent-slot-type/{slot_type_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Mount API

    async def get_mounts_index(self):
        """Returns an index of mounts.

        :return: Returns an index of mounts.
        :rtype: dict
        """
        endpoint = f"/data/wow/mount/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mount(self, mount_id: int):
        """Returns a mount by ID.

        :param mount_id: The ID of the mount.
        :type mount_id: int
        :return: Returns a mount by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/mount/{mount_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mount_search(self, search_params: Optional[dict] = None):
        """Performs a search of mounts. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of mounts. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/mount"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Mythic Keystone Affix API

    async def get_mythic_keystone_affixes_index(self):
        """Returns an index of mythic keystone affixes.

        :return: Returns an index of mythic keystone affixes.
        :rtype: dict
        """
        endpoint = f"/data/wow/keystone-affix/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_affix(self, keystone_affix_id: int):
        """Returns a mythic keystone affix by ID.

        :param keystone_affix_id: The ID of the mythic keystone affix.
        :type keystone_affix_id: int
        :return: Returns a mythic keystone affix by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/keystone-affix/{keystone_affix_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_affix_media(self, keystone_affix_id: int):
        """Returns media for a mythic keystone affix by ID.

        :param keystone_affix_id: The ID of the mythic keystone affix.
        :type keystone_affix_id: int
        :return: Returns media for a mythic keystone affix by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/keystone-affix/{keystone_affix_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Mythic Keystone Dungeon API

    async def get_mythic_keystone_dungeons_index(self):
        """Returns an index of Mythic Keystone dungeons.

        :return: Returns an index of Mythic Keystone dungeons.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/dungeon/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_dungeon(self, dungeon_id: int):
        """Returns a Mythic Keystone dungeon by ID.

        :param dungeon_id: The ID of the dungeon.
        :type dungeon_id: int
        :return: Returns a Mythic Keystone dungeon by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/dungeon/{dungeon_id}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_index(self):
        """Returns an index of links to other documents related to Mythic Keystone dungeons.

        :return: Returns an index of links to other documents related to Mythic Keystone dungeons.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_periods_index(self):
        """Returns an index of Mythic Keystone periods.

        :return: Returns an index of Mythic Keystone periods.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/period/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_period(self, period_id: int):
        """Returns a Mythic Keystone period by ID.

        :param period_id: The ID of the Mythic Keystone season period.
        :type period_id: int
        :return: Returns a Mythic Keystone period by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/period/{period_id}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_seasons_index(self):
        """Returns an index of Mythic Keystone seasons.

        :return: Returns an index of Mythic Keystone seasons.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/season/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_season(self, season_id: int):
        """Returns a Mythic Keystone season by ID.

        :param season_id: The ID of the Mythic Keystone season.
        :type season_id: int
        :return: Returns a Mythic Keystone season by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/mythic-keystone/season/{season_id}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Mythic Keystone Leaderboard API

    async def get_mythic_keystone_leaderboards_index(self, connected_realm_id: int):
        """Returns an index of Mythic Keystone Leaderboard dungeon instances for a connected realm.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :return: Returns an index of Mythic Keystone Leaderboard dungeon instances for a connected realm.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}/mythic-leaderboard/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_mythic_keystone_leaderboard(self, connected_realm_id: int, dungeon_id: int, period: int):
        """Returns a weekly Mythic Keystone Leaderboard by period.

        :param connected_realm_id: The ID of the connected realm.
        :type connected_realm_id: int
        :param dungeon_id: The ID of the dungeon.
        :type dungeon_id: int
        :param period: The unique identifier for the leaderboard period.
        :type period: int
        :return: Returns a weekly Mythic Keystone Leaderboard by period.
        :rtype: dict
        """
        endpoint = f"/data/wow/connected-realm/{connected_realm_id}/mythic-leaderboard/{dungeon_id}/period/{period}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Mythic Raid Leaderboard API

    async def get_mythic_raid_leaderboard(self, raid: str, faction: str):
        """Returns the leaderboard for a given raid and faction.

        :param raid: The raid for a leaderboard.
        :type raid: str
        :param faction: Player faction (alliance or horde).
        :type faction: str
        :return: Returns the leaderboard for a given raid and faction.
        :rtype: dict
        """
        endpoint = f"/data/wow/leaderboard/hall-of-fame/{raid}/{faction}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Pet API

    async def get_pets_index(self):
        """Returns an index of battle pets.

        :return: Returns an index of battle pets.
        :rtype: dict
        """
        endpoint = f"/data/wow/pet/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pet(self, pet_id: int):
        """Returns a battle pets by ID.

        :param pet_id: The ID of the pet.
        :type pet_id: int
        :return: Returns a battle pets by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pet/{pet_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pet_media(self, pet_id: int):
        """Returns media for a battle pet by ID.

        :param pet_id: The ID of the pet.
        :type pet_id: int
        :return: Returns media for a battle pet by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/pet/{pet_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pet_abilities_index(self):
        """Returns an index of pet abilities.

        :return: Returns an index of pet abilities.
        :rtype: dict
        """
        endpoint = f"/data/wow/pet-ability/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pet_ability(self, pet_ability_id: int):
        """Returns a pet ability by ID.

        :param pet_ability_id: The ID of the pet ability.
        :type pet_ability_id: int
        :return: Returns a pet ability by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pet-ability/{pet_ability_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pet_ability_media(self, pet_ability_id: int):
        """Returns media for a pet ability by ID.

        :param pet_ability_id: The ID of the pet ability.
        :type pet_ability_id: int
        :return: Returns media for a pet ability by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/pet-ability/{pet_ability_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Playable Class API

    async def get_playable_classes_index(self):
        """Returns an index of playable classes.

        :return: Returns an index of playable classes.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-class/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_class(self, class_id: int):
        """Returns a playable class by ID.

        :param class_id: The ID of the playable class.
        :type class_id: int
        :return: Returns a playable class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-class/{class_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_class_media(self, playable_class_id: int):
        """Returns media for a playable class by ID.

        :param playable_class_id: The ID of the playable class.
        :type playable_class_id: int
        :return: Returns media for a playable class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/playable-class/{playable_class_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_talent_slots(self, class_id: int):
        """Returns the PvP talent slots for a playable class by ID.

        :param class_id: The ID of the playable class.
        :type class_id: int
        :return: Returns the PvP talent slots for a playable class by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-class/{class_id}/pvp-talent-slots"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Playable Race API

    async def get_playable_races_index(self):
        """Returns an index of playable races.

        :return: Returns an index of playable races.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-race/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_race(self, playable_race_id: int):
        """Returns a playable race by ID.

        :param playable_race_id: The ID of the playable race.
        :type playable_race_id: int
        :return: Returns a playable race by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-race/{playable_race_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Playable Specialization API

    async def get_playable_specializations_index(self):
        """Returns an index of playable specializations.

        :return: Returns an index of playable specializations.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-specialization/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_specialization(self, spec_id: int):
        """Returns a playable specialization by ID.

        :param spec_id: The ID of the playable specialization.
        :type spec_id: int
        :return: Returns a playable specialization by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/playable-specialization/{spec_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_playable_specialization_media(self, spec_id: int):
        """Returns media for a playable specialization by ID.

        :param spec_id: The ID of the playable specialization.
        :type spec_id: int
        :return: Returns media for a playable specialization by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/playable-specialization/{spec_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Power Type API

    async def get_power_types_index(self):
        """Returns an index of power types.

        :return: Returns an index of power types.
        :rtype: dict
        """
        endpoint = f"/data/wow/power-type/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_power_type(self, power_type_id: int):
        """Returns a power type by ID.

        :param power_type_id: The ID of the power type.
        :type power_type_id: int
        :return: Returns a power type by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/power-type/{power_type_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Profession API

    async def get_professions_index(self):
        """Returns an index of professions.

        :return: Returns an index of professions.
        :rtype: dict
        """
        endpoint = f"/data/wow/profession/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_profession(self, profession_id: int):
        """Returns a profession by ID.

        :param profession_id: The ID of the profession.
        :type profession_id: int
        :return: Returns a profession by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/profession/{profession_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_profession_media(self, profession_id: int):
        """Returns media for a profession by ID.

        :param profession_id: The ID of the profession.
        :type profession_id: int
        :return: Returns media for a profession by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/profession/{profession_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_profession_skill_tier(self, profession_id: int, skill_tier_id: int):
        """Returns a skill tier for a profession by ID.

        :param profession_id: The ID of the profession.
        :type profession_id: int
        :param skill_tier_id: The ID of the skill tier.
        :type skill_tier_id: int
        :return: Returns a skill tier for a profession by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/profession/{profession_id}/skill-tier/{skill_tier_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_recipe(self, recipe_id: int):
        """Returns a recipe by ID.

        :param recipe_id: The ID of the recipe.
        :type recipe_id: int
        :return: Returns a recipe by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/recipe/{recipe_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_recipe_media(self, recipe_id: int):
        """Returns media for a recipe by ID.

        :param recipe_id: The ID of the recipe.
        :type recipe_id: int
        :return: Returns media for a recipe by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/recipe/{recipe_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region PvP Season API

    async def get_pvp_seasons_index(self):
        """Returns an index of PvP seasons.

        :return: Returns an index of PvP seasons.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_season(self, pvp_season_id: int):
        """Returns a PvP season by ID.

        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns a PvP season by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/{pvp_season_id}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_leaderboards_index(self, pvp_season_id: int):
        """Returns an index of PvP leaderboards for a PvP season.

        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns an index of PvP leaderboards for a PvP season.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/{pvp_season_id}/pvp-leaderboard/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_leaderboard(self, pvp_season_id: int, pvp_bracket: str):
        """Returns the PvP leaderboard of a specific PvP bracket for a PvP season.

        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :param pvp_bracket: The PvP bracket type.
        :type pvp_bracket: str
        :return: Returns the PvP leaderboard of a specific PvP bracket for a PvP season.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/{pvp_season_id}/pvp-leaderboard/{pvp_bracket}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_rewards_index(self, pvp_season_id: int):
        """Returns an index of PvP rewards for a PvP season.

        :param pvp_season_id: The ID of the PvP season.
        :type pvp_season_id: int
        :return: Returns an index of PvP rewards for a PvP season.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-season/{pvp_season_id}/pvp-reward/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region PvP Tier API

    async def get_pvp_tier_media(self, pvp_tier_id: int):
        """Returns media for a PvP tier by ID.

        :param pvp_tier_id: The ID of the PvP tier.
        :type pvp_tier_id: int
        :return: Returns media for a PvP tier by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/pvp-tier/{pvp_tier_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_tiers_index(self):
        """Returns an index of PvP tiers.

        :return: Returns an index of PvP tiers.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-tier/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_tier(self, pvp_tier_id: int):
        """Returns a PvP tier by ID.

        :param pvp_tier_id: The ID of the PvP tier.
        :type pvp_tier_id: int
        :return: Returns a PvP tier by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-tier/{pvp_tier_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Quest API

    async def get_quests_index(self):
        """Returns the parent index for quests.

        :return: Returns the parent index for quests.
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest(self, quest_id: int):
        """Returns a quest by ID.

        :param quest_id: The ID of the quest.
        :type quest_id: int
        :return: Returns a quest by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/{quest_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest_categories_index(self):
        """Returns an index of quest categories (such as quests for a specific class, profession, or storyline).

        :return: Returns an index of quest categories (such as quests for a specific class, profession, or storyline).
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/category/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest_category(self, quest_category_id: str):
        """Returns a quest category by ID.

        :param quest_category_id: The ID of the quest category.
        :type quest_category_id: str
        :return: Returns a quest category by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/category/{quest_category_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest_areas_index(self):
        """Returns an index of quest areas.

        :return: Returns an index of quest areas.
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/area/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest_area(self, quest_area_id: str):
        """Returns a quest area by ID.

        :param quest_area_id: The ID of the quest area.
        :type quest_area_id: str
        :return: Returns a quest area by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/area/{quest_area_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest_types_index(self):
        """Returns an index of quest types (such as PvP quests, raid quests, or account quests).

        :return: Returns an index of quest types (such as PvP quests, raid quests, or account quests).
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/type/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_quest_type(self, quest_type_id: str):
        """Returns a quest type by ID.

        :param quest_type_id: The ID of the quest type.
        :type quest_type_id: str
        :return: Returns a quest type by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/quest/type/{quest_type_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Realm API

    async def get_realms_index(self):
        """Returns an index of realms.

        :return: Returns an index of realms.
        :rtype: dict
        """
        endpoint = f"/data/wow/realm/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_realm(self, realm_slug: str):
        """Returns a single realm by slug or ID.

        :param realm_slug: The slug of the realm.
        :type realm_slug: str
        :return: Returns a single realm by slug or ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/realm/{realm_slug}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_realm_search(self, search_params: Optional[dict] = None):
        """Performs a search of realms. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of realms. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/realm"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Region API

    async def get_regions_index(self):
        """Returns an index of regions.

        :return: Returns an index of regions.
        :rtype: dict
        """
        endpoint = f"/data/wow/region/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_region(self, region_id: int):
        """Returns a region by ID.

        :param region_id: The ID of the region.
        :type region_id: int
        :return: Returns a region by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/region/{region_id}"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Reputations API

    async def get_reputation_factions_index(self):
        """Returns an index of reputation factions.

        :return: Returns an index of reputation factions.
        :rtype: dict
        """
        endpoint = f"/data/wow/reputation-faction/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_reputation_faction(self, reputation_faction_id: int):
        """Returns a single reputation faction by ID.

        :param reputation_faction_id: The ID of the reputation faction.
        :type reputation_faction_id: int
        :return: Returns a single reputation faction by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/reputation-faction/{reputation_faction_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_reputation_tiers_index(self):
        """Returns an index of reputation tiers.

        :return: Returns an index of reputation tiers.
        :rtype: dict
        """
        endpoint = f"/data/wow/reputation-tiers/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_reputation_tiers(self, reputation_tiers_id: int):
        """Returns a single set of reputation tiers by ID.

        :param reputation_tiers_id: The ID of the set of reputation tiers.
        :type reputation_tiers_id: int
        :return: Returns a single set of reputation tiers by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/reputation-tiers/{reputation_tiers_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Spell API

    async def get_spell(self, spell_id: int):
        """Returns a spell by ID.

        :param spell_id: The ID of the spell.
        :type spell_id: int
        :return: Returns a spell by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/spell/{spell_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_spell_media(self, spell_id: int):
        """Returns media for a spell by ID.

        :param spell_id: The ID of the spell.
        :type spell_id: int
        :return: Returns media for a spell by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/spell/{spell_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_spell_search(self, search_params: Optional[dict] = None):
        """Performs a search of spells. The fields below are provided for example. For more detail see the Search Guide.

        :param search_params: Search parameters
        :type search_params: dict
        :return: Performs a search of spells. The fields below are provided for example. For more detail see the Search Guide.
        :rtype: dict
        """
        endpoint = f"/data/wow/search/spell"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint, params=search_params)

    # endregion
    # region Talent API

    async def get_talents_index(self):
        """Returns an index of talents.

        :return: Returns an index of talents.
        :rtype: dict
        """
        endpoint = f"/data/wow/talent/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_talent(self, talent_id: int):
        """Returns a talent by ID.

        :param talent_id: The ID of the talent.
        :type talent_id: int
        :return: Returns a talent by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/talent/{talent_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_talents_index(self):
        """Returns an index of PvP talents.

        :return: Returns an index of PvP talents.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-talent/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_pvp_talent(self, pvp_talent_id: int):
        """Returns a PvP talent by ID.

        :param pvp_talent_id: The ID of the PvP talent.
        :type pvp_talent_id: int
        :return: Returns a PvP talent by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/pvp-talent/{pvp_talent_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Tech Talent API

    async def get_tech_talent_tree_index(self):
        """Returns an index of tech talent trees.

        :return: Returns an index of tech talent trees.
        :rtype: dict
        """
        endpoint = f"/data/wow/tech-talent-tree/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_tech_talent_tree(self, tech_talent_tree_id: int):
        """Returns a tech talent tree by ID.

        :param tech_talent_tree_id: The ID of the tech talent tree.
        :type tech_talent_tree_id: int
        :return: Returns a tech talent tree by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/tech-talent-tree/{tech_talent_tree_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_tech_talent_index(self):
        """Returns an index of tech talents.

        :return: Returns an index of tech talents.
        :rtype: dict
        """
        endpoint = f"/data/wow/tech-talent/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_tech_talent(self, tech_talent_id: int):
        """Returns a tech talent by ID.

        :param tech_talent_id: The ID of the tech talent.
        :type tech_talent_id: int
        :return: Returns a tech talent by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/tech-talent/{tech_talent_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_tech_talent_media(self, tech_talent_id: int):
        """Returns media for a tech talent by ID.

        :param tech_talent_id: The ID of the tech talent.
        :type tech_talent_id: int
        :return: Returns media for a tech talent by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/media/tech-talent/{tech_talent_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region Title API

    async def get_titles_index(self):
        """Returns an index of titles.

        :return: Returns an index of titles.
        :rtype: dict
        """
        endpoint = f"/data/wow/title/index"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_title(self, title_id: int):
        """Returns a title by ID.

        :param title_id: The ID of the title.
        :type title_id: int
        :return: Returns a title by ID.
        :rtype: dict
        """
        endpoint = f"/data/wow/title/{title_id}"
        namespace = "static-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
    # region WoW Token API

    async def get_wow_token_index(self):
        """Returns the WoW Token index.

        :return: Returns the WoW Token index.
        :rtype: dict
        """
        endpoint = f"/data/wow/token/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    async def get_wow_token_index_cn(self):
        """Returns the WoW Token index.

        :return: Returns the WoW Token index.
        :rtype: dict
        """
        endpoint = f"/data/wow/token/index"
        namespace = "dynamic-{region}"

        return await self.get_game_api_resource(namespace, endpoint)

    # endregion
